/*ToDo:


*/
/*
Realisiert mit AT-Mega88, 16MHz

Ausgnge:
PD1     TX zum LCD


Eingnge:
PC0     Spannung
PC1     Strom
PC2     Offset
PB0     Nulldurchgang
PD0     Offsetmessung
PD2     Schalterstellung 400mA
PD3     Schalterstellung 100mA
PD4     Schalterstellung  25mA


==========================================
T0 - 8Bit:   Messung von Verweildauer ausserhalb des ADC-Interrupts
T1 - 16Bit:  Phasenwinkelmessung
T2 - 8bit:   Zeitbasis fr Delay-Funktion



Normierung Spannung:
  Theorie:
    230 Veff sind 325 Vmax
    Mit Spannungsteiler 400k/2K7 gibt das 2.18V Spitzenspannung
  
    Mit 5V Uref und 1024 Schritten des A/D Wandlers:
    1 Schritt sind 5/1024 V
    2.18V = 446 Schritte
    446/325=1.372
  
    Spannung: Schritte / 1.372

Normierung Strom:
  400mA geben 2.5V
  1 Schritt sind 5/1024 V
  400mA = 512 Schritte
  512/400=1.28
  
  Strom: Schritte / 1.28

  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  Praxis:

    219V Netzspannung gemessen. Gibt 310V Scheitelspannung.
      1,04V -> 310V 
    ADC zeigt 512+212.
      212Schritte -> 310V
      U = Schritte * 1,462
  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


!!! ACHTUNG !!!
Fr Floating-Point Ausgabe folgende Libraries mit einbinden:
  libm.a
  libprintf_flt.a

und als Custom-Option fr den Linker
  -Wl,-u,vfprintf
eintragen.

siehe:
http://winavr.scienceprog.com/avr-gcc-tutorial/using-sprintf--function-for-float-numbers-in-avr-gcc.html


*/



#define F_CPU 16000000L	/* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h> //sprintf
#include <math.h>  //sqrt

typedef unsigned char byte;

#define U_KALIBRIER 1462L
#define I_KALIBRIER 1462L
#define PHI_KALIBRIER 3466L

//Der Einfachheit halber: viele globale Variablen

//Wir haben insgesamt 1024 Bytes SRAM - reichlich Platz, um alle Variablen statisch zu halten
typedef struct {
  int offsetU;
  int offsetI;
  unsigned int countU;
  unsigned int countI;
  unsigned int countPhi;
  long sumSpannung;
  long sumStrom;
  long sumWirk;
  long winkel;
  long sumWinkel;
//  byte modus;
  volatile byte messen;
  volatile byte verweildauer;
  volatile byte verweildauerMax;
  volatile byte verweildauerMin;
  volatile unsigned int ms;
  volatile byte finished;
  int LemWindungen;
    
  long Umax;
  long Imax;

}Tglob;

//Alle globalen Var in einer Struct deklarieren
Tglob g;


//--------------------------------------------------------
//--------------------------------------------------------
//--------------------------------------------------------
//--------------------------------------------------------
//--------------------------------------------------------
// ms - Timer
ISR(TIMER2_COMPA_vect){ 
  g.ms--;
  if (0==g.ms){  //Bewertung muss innerhalb der IRQ-Routine erfolgen,
                 //da Bewertung der 16Bit Variable sonst unterbrochen 
                 //werden kann!
    g.finished = 1;
  }
}
//--------------------------------------------------------
//--------------------------------------------------------

//ADC-Ergebnis fertig
ISR(ADC_vect){
  //ADC liefert mit 9KHz einen Messwert
  static byte channel=0;
  static byte nextChannel=0;
  static byte TriggerUscharf;
  static byte TriggerIscharf;
  static long letzteSpannung;
  static int messwert;
  static long istwert;

  g.verweildauer = TCNT0;
  PORTD|=1;

  messwert = ADC;
  channel = ADMUX & 3;
  ADMUX = _BV(REFS0)|nextChannel;
  ADCSRA |= _BV(ADSC); //Start next conversion
  
  if (g.messen){
  
    if (g.verweildauerMax < g.verweildauer){
      g.verweildauerMax = g.verweildauer;
    }
    if (g.verweildauerMin > g.verweildauer){
      g.verweildauerMin = g.verweildauer;
    }

    if (0==channel){ //Spannung
      istwert = ((messwert - g.offsetU) * U_KALIBRIER)/1000L;
      letzteSpannung = istwert;
      g.sumSpannung += istwert * istwert;
      g.countU++;
      if (g.Umax < istwert){
        g.Umax = istwert;
      }
      if (istwert < -40) TriggerUscharf=1;
        
      if (TriggerUscharf && (istwert > 0)){
        //positiven Nulldurchgang der Spannung entdeckt
        //Timer T1 lschen und starten

        TCCR1B =  0; //Stopp timer
        TCNT1=0;
        TCCR1B =  _BV(CS12); // Starte Timer mit Vorteiler 256
        TriggerUscharf=0;
      }
    }else
    if (1==channel){ //Strom
      istwert = ((messwert - g.offsetI)  * I_KALIBRIER)/1000L;
      g.sumStrom += istwert * istwert;
      g.countI++;
      g.sumWirk += letzteSpannung * istwert;
      if (g.Imax < istwert){
        g.Imax = istwert;
      }
      
      if (istwert < -4) TriggerIscharf=1;
      if (TriggerIscharf && (istwert > 0) && TCCR1B){
        //positiven Nulldurchgang des Stromes entdeckt
        //Timer T1 auswerten
        TCCR1B =  0; //Stopp timer
        g.winkel = TCNT1;
        TriggerIscharf=0;
        g.sumWinkel += g.winkel;
        g.countPhi++;
      }
    }
  }
  
  if (nextChannel){
    nextChannel=0;
  }else{
    nextChannel=1;
  }
  PORTD&=~1;
  TCNT0 = 0;
}
//--------------------------------------------------------
//--------------------------------------------------------
void offsetUMessung(void){
  byte i;
  int sum;
  ADMUX = _BV(REFS0)|2; //Internal 5V Ref; Right adjust; MUX=Ch2 (Offset-Kanal)
  ADCSRA = _BV(ADEN)| //Enable
          _BV(ADSC) | //Start conversion
          _BV(ADATE)| //Auto trigger (free running )
          _BV(ADPS2)| //Prescaler=128  
          _BV(ADPS1)|
          _BV(ADPS0);
  ADCSRB = 0; //free running

  ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
  while(!(ADCSRA & _BV(ADIF)))
  ;
  sum=0;
  for (i=0; i<32; i++){
    ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
    while(!(ADCSRA & _BV(ADIF))) //Wait for conversion complete
    ;
    sum+=ADC;  //ADCL + 256*ADCH
  }
  g.offsetU = sum / 32;
  ADCSRA = _BV(ADEN)| //Enable
          //_BV(ADSC) | //Start conversion
          //_BV(ADATE) | //Free running off
          _BV(ADIE) | //Interrupt ein
          _BV(ADPS2)| //Prescaler=128  
          _BV(ADPS1)|
          _BV(ADPS0);
  ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it


  //Vorlufig: Offset I = Offset U
  //Eigentlich muss bei I=0 auch der Offset des LEM mitgemessen werden.
  //Dazu msste ein Relais rein, welches I unterbricht.
  g.offsetI = g.offsetU;
  
}
//--------------------------------------------------------------------------
void offsetIMessung(void){
  byte i;
  int sum;
  ADMUX = _BV(REFS0)|1; //Internal 5V Ref; Right adjust; MUX=Ch1 (Strom)
  ADCSRA = _BV(ADEN)| //Enable
          _BV(ADSC) | //Start conversion
          _BV(ADATE)| //Auto trigger (free running )
          _BV(ADPS2)| //Prescaler=128  
          _BV(ADPS1)|
          _BV(ADPS0);
  ADCSRB = 0; //free running

  ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
  while(!(ADCSRA & _BV(ADIF)))
  ;
  sum=0;
  for (i=0; i<32; i++){
    ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
    while(!(ADCSRA & _BV(ADIF))) //Wait for conversion complete
    ;
    sum+=ADC;  //ADCL + 256*ADCH
  }
  g.offsetI = sum / 32;
  ADCSRA = _BV(ADEN)| //Enable
          //_BV(ADSC) | //Start conversion
          //_BV(ADATE) | //Free running off
          _BV(ADIE) | //Interrupt ein
          _BV(ADPS2)| //Prescaler=128  
          _BV(ADPS1)|
          _BV(ADPS0);
  ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it

}



//--------------------------------------------------------
void delay_ms(int dauer_ms){
#if 1
  //Set up the 8-bit timer 2  (1KHz Generator)
  //------------------------------------------
  TCCR2B = 0; //AUS. 5=Teiler durch 128
  TCCR2A = (1<<WGM21); //CTC _Modus
  OCR2A = 124; //Teilt durch 125

  //Timer luft mit 1 KHz ber 


  //Zhler reseten
  TCNT2 = 0;
  g.ms=dauer_ms;
  g.finished=0;

  //Timer starten
  TIMSK2 = (1<<OCIE2A);
  TCCR2B = 5;  //5= Teiler durch 128

  
  //Warten, bis Timer Endwert erreicht
  while (!g.finished)
  ; 

  
  //Timer 2 abschalten
  TCCR2B = 0;
  TIMSK2 = 0;
  
#else   

//Notlsung ohne Timer. Handkalibriert fr -O0.
int i,j;

  for (i=0; i<dauer_ms; i++){
    for (j=0; j<613; j++){
   
      asm volatile ("push r0");
      asm volatile ("pop r0");
      asm volatile ("NOP");
      asm volatile ("nop");

    }
  }
#endif
}
//--------------------------------------------------------
void USART_Transmit( unsigned char data ){
//--------------------------------------------------------
	/* Wait for empty transmit buffer */
	while ( !( UCSR0A & (1<<UDRE0)) )
	;
	/* Put data into buffer, sends the data */
	UDR0 = data;
}
//--------------------------------------------------------
void sendstringRS232(unsigned char *text){
//--------------------------------------------------------
  while(*text){
    USART_Transmit(*text++);
  }
}


//--------------------------------------------------------

int main (void)
{
int i;
byte *ram;
long signed int phi;
double ueff,ieff,wirk;
byte buffer[140];
byte halbwelle;
byte halbwellenZaehler;
byte schalter;
byte ueberlast;

    cli();

    //----- Disable the watchdog ----------------------------------
    asm("wdr"); //__watchdog_reset();
    /* Clear WDRF in MCUSR */
    MCUSR &= ~(1<<WDRF);
    /* Write logical one to WDCE and WDE */
    /* Keep old prescaler setting to prevent unintentional time-out */
    WDTCSR |= (1<<WDCE) | (1<<WDE);
    /* Turn off WDT */
    WDTCSR = 0x00;


    //---- Set port directions ----------------------------------
    DDRC = 8+16+32;
    PORTC = 0; //alles AUS

    DDRD = 0; //Alles Eingnge (TX wird autom. zum Ausgang)
    PORTD = 255;  //alle Pull-Up Widerstnde ein



    DDRB = 0;
    PORTB = 0; //keine Pull-Ups.

    //----- setup A/D converter  --------------------------------
    //ADC - single Conversion Mode (not Free Running)
    ADMUX = _BV(REFS0); //Internal 5V Ref; Right adjust(10Bit); MUX=Ch0
    ADCSRA = _BV(ADEN)| _BV(ADIE)|_BV(ADPS2)|_BV(ADPS1)|_BV(ADPS0);//Enable; IRQ Enable; Prescaler128 
                                                                   //  gibt bei 16MHz: 125KHz ADC
    //ADC liefert mit knapp 9KHz einen Messwert 

         //Set up the 8-bit timer 0 (500ns Takt)  - Messung von Verweildauer
    TCCR0A = 0;          //Normaler Timer/Counter;
    TCCR0B = _BV(CS01) ; // Prescaler= /8 (gibt 2MHz)
    TCNT0=0;
    //TIFR0 |=1; //Lscht (!) TOV0 Bit

        //Set up the 16-bit timer 1 (winkelmessung).
    TCCR1A =  0; //Normaler Timer/Counter
    TCCR1B =  0;//Spter:  _BV(CS12); // Vorteiler 256
    //Zeitbasis: 62500 KHz 16us 
    //Zhlt max. bis 65535 - also ca. 1 Sekunde
    //Bei L=90 (5ms): Zhlerstand=312
    //Gradzahl = Zhlerstand /3,466666 
    //Bei L=-0 (20ms): Zhlerstand=1250
    


    //----- RS232 initialisieren --------------------------------------------------

    /* Set baud rate */
    UBRR0H = 0;
    UBRR0L = 103;  //9600 Baud bei 16MHz Takt
    UCSR0A=0; //Clear U2X
  
    /* Enable transmitter, no Interrupts*/
    UCSR0B = (1<<TXEN0);
  
    /* Set frame format: 8data, 1stop bit, no parity */
    UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);


    
    //Globale Variablen mit 0 initialisieren
    ram=(unsigned char *)&g;
    for (i=0; i< sizeof(Tglob); ram[i++]=0); 


    sei(); //Interrupts EIN. Jetzt geht's los!


//while (1){PORTD ^=255 ; delay_ms(1000);};



    g.messen=0;
    for (i=3; i>0; i--){
      delay_ms(250);
      sendstringRS232("|o|*|u|*|1Wirkleistungsmessgeraet");

      delay_ms(10);
      sendstringRS232("|3D. Schroeder 01/2009");

      delay_ms(10);
      sprintf(buffer,"|4 %d  ",i); 
      sendstringRS232(buffer);
    }
    //sendstringRS232("|o|*|u|*|2 Offset gemessen.");

    offsetIMessung();
    offsetUMessung();

    //------------- Hauptschleife ------------------------------ 
    while(1){

      //Bereichsschalter auslesen
      schalter = PIND;
      if (!(schalter & 1)){

        offsetIMessung();
        sprintf(buffer,"|1|*Strom Offsetmessung|2I-Offset=%d|u|*",g.offsetI); 
        sendstringRS232(buffer);
        delay_ms(333);
        continue;
      }else  
      if (!(schalter & 4)){
        g.LemWindungen=1;  //400mA 92W
      }else  
      if (!(schalter & 8)){
        g.LemWindungen=4;  //100mA 23W
      }else  
      if (!(schalter & 16)){
        g.LemWindungen=16; //25mA 5,75W
      }else{                      //123456789 123456789 1234567
        sendstringRS232("|o|*|u|*|2     <- weiter drehen ->");
        delay_ms(333);
        continue;
      }

      g.verweildauerMax =   0;
      g.verweildauerMin = 255;
      g.sumSpannung =0;
      g.sumStrom=0;
      g.sumWirk=0;
      g.sumWinkel=0;
      g.countU=0;
      g.countI=0;
      g.countPhi=0;
      g.Umax=0;
      g.Imax=0;
      halbwellenZaehler = 0;

      //Warten auf Start einer Vollwelle (Nulldurchgang)
      halbwelle =PINB & 1;
      while (halbwelle ==(PINB & 1)){
      };

      ADCSRA |= _BV(ADSC); //Start next conversion

      g.messen=1;
      halbwelle =PINB & 1;
      //Messe bis gengend Halbwellen durchgelaufen
      do {
        if (halbwelle !=(PINB & 1)){
          halbwelle =PINB & 1;
          halbwellenZaehler++;
        }
      }while (halbwellenZaehler < 16);
  
        
      //Auswertung
      g.messen=0;
      ueff = sqrt(g.sumSpannung / g.countU); //in V
      ieff = sqrt(g.sumStrom / g.countI);     //in mA 
      if (ieff > 400){
        ueberlast=1;
      }else{
        ueberlast=0;
      }
      ieff /= g.LemWindungen;
      wirk = g.sumWirk / g.countI;
      wirk /= g.LemWindungen;
      wirk /=1000;
      phi  = g.sumWinkel / g.countPhi;

      phi=(phi*1000)/PHI_KALIBRIER;
      if (phi > 180) phi -= 360;

//      sprintf(buffer,"|1|*Ueff=%3.0fV  Ieff=%#5.1fmA|2P=%6.3fW phi=%4ld ",ueff,ieff,wirk,phi); 
      if (g.countPhi){
        sprintf(buffer,"|1|*P=%#7.3fW phi=%4ld |2Ueff=%3.0fV  Ieff=%#5.1fmA",wirk,phi,ueff,ieff); 
      }else{
        sprintf(buffer,"|1|*P=%#7.3fW phi= -?-  |2Ueff=%3.0fV  Ieff=%#5.1fmA",wirk,ueff,ieff); 
      }
      sendstringRS232(buffer);

//      sprintf(buffer,"|3|*max=%ldV %ldmA", g.Umax,g.Imax); 
      switch(g.LemWindungen){
        case 1:
          sprintf(buffer,"|3|*Bereich  25  100 [400]mA "); 
          break;
        case 4:
          sprintf(buffer,"|3|*Bereich  25 [100] 400 mA "); 
//        sprintf(buffer,"|3|*Bereich 0..100mA 0 - 23W"); 
          break;
        case 16:
          sprintf(buffer,"|3|*Bereich [25] 100  400 mA "); 
//        sprintf(buffer,"|3|*Bereich 0..25mA 0 - 5.8W"); 
          break;
      }
      //sprintf(buffer,"|3|*I;Uoffs=%d;%d LEM=%dWdg", g.offsetI,g.offsetU, g.LemWindungen); 
      sendstringRS232(buffer);
      if (ueberlast){
                         //123456789 123456789 1234567
        sprintf(buffer, "|4    ---  Ueberlast  ---"); 

      }else{
//        sprintf(buffer,"|4Nphi=%d Tmin=%d Ni=%d ",g.countPhi, g.verweildauerMin, g.countI); 
        sprintf(buffer,"|4Ni=%d Nphi=%d Ioffs=%d", g.countI, g.countPhi, g.offsetI); 
      }
      sendstringRS232(buffer);


      delay_ms(180);

    }//while

    return 0;
}
